@@ -2,13 +2,14 @@ from __future__ import annotations
 
 import json
 import re
+import warnings
 from hashlib import md5
 from logging import WARNING, DEBUG, INFO, Logger
 from pathlib import Path
 from typing import Optional, Generic, Any, Dict, cast, TypeVar, Iterable
 
 from persistable.data import PersistableParams
-from persistable.exceptions import ExplainedNotImplementedError
+from persistable.exceptions import ExplainedNotImplementedError, NoPayloadError, InvalidPayloadWarning
 from persistable.io import FileIO, PickleFileIO, PayloadTypeT
 from persistable.logging import get_logger
 
@@ -96,21 +97,19 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
         self.logger.info(f"Payload named {payload_name}; Parameters set to {params}")
 
         self._payload: Optional[PayloadTypeT] = None
-        self._params_tree: Optional[Dict[str, Any]] = None
 
     @property
     def params_tree(self) -> Dict[str, Any]:
-        if self._params_tree is None:
-            self._params_tree = self.params.to_dict() | {
-                persistable_obj.payload_name: persistable_obj.params.to_dict()
-                for persistable_obj in self.tracked_persistable_dependencies
-            }
-
-        return self._params_tree
+        return self.params.to_dict() | {
+            persistable_obj.payload_name: persistable_obj.params_tree
+            for persistable_obj in self.tracked_persistable_dependencies
+        }
 
     def generate(self, persist: bool = True, **untracked_payload_params: Any) -> None:
         """
-        Generates payload and (by default) persist it.
+        Generate payload and (by default) persist it.
+
+        This method requires an implementation of _generate_payload().
 
         Parameters
         ----------
@@ -120,12 +119,12 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
             These are helper parameters for generating an object that are not tracked.
             Generally these are not used.
 
-        Returns
-        -------
 
         """
         self.logger.info(f"Now generating {self.payload_name} payload...")
-        self._payload = self._generate_payload(**untracked_payload_params)
+        generated_payload = self._generate_payload(**untracked_payload_params)
+        self._validate_payload(payload=generated_payload)
+        self._payload = generated_payload
         if persist:
             self.persist()
 
@@ -136,19 +135,25 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
         if self._payload is None:
             raise ValueError("Payload has not been generated.")
 
-        self.payload_io.save(
-            payload=self._payload,
-            filepath=self.persist_filepath.with_suffix(self.payload_file_suffix),
-        )
-        with self.persist_filepath.with_suffix(".params.json").open("w") as params_file_handler:
-            json.dump(self.params_tree, params_file_handler)
+        payload_filepath = self.persist_filepath.with_suffix(self.payload_file_suffix)
+        params_filepath = self.persist_filepath.with_suffix(".params.json")
 
-    def load(self, **untracked_payload_params: Any) -> None:
+        self.payload_io.save(payload=self._payload, filepath=payload_filepath)
+        with params_filepath.open("w") as params_file_handler:
+            json.dump(self.params_tree, params_file_handler)
+        self.logger.info(
+            f"Successfully persisted payload to {payload_filepath.name} "
+            f"(see {params_filepath.name} to view corresponding params)."
+        )
+
+    def load(self, warn_if_validation_fails: bool = False, **untracked_payload_params: Any) -> None:
         """
         Loads persisted payload
 
         Parameters
         ----------
+        warn_if_validation_fails    : bool
+            Use True to catch exceptions from valid_payload and throw a warning instead.
         untracked_payload_params    : dict
             Parameters not tracked by persistable that are only used to run the _post_load.
             Such scripts are useful if part of the payload cannot be persisted and needs to be recalculated
@@ -160,7 +165,23 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
         """
         self.logger.info(f"Now loading {self.payload_name} payload...")
         # ToDo - add find similar file functionality
-        self._payload = self.payload_io.load(self.persist_filepath.with_suffix(self.payload_file_suffix))
+
+        payload_filepath = self.persist_filepath.with_suffix(self.payload_file_suffix)
+        loaded_payload = self.payload_io.load(filepath=payload_filepath)
+
+        try:
+            self._validate_payload(loaded_payload)
+        except Exception as exception:
+            if warn_if_validation_fails:
+                warning_message = "Loaded payload was invalid (it may need to be regenerated)"
+                warnings.warn(message=warning_message, category=InvalidPayloadWarning)
+                self.logger.warning(msg=warning_message, exc_info=exception)
+            else:
+                raise exception
+
+        self._payload = loaded_payload
+        self.logger.info(f"Successfully loaded payload from {payload_filepath.name}")
+
         self._post_load()
 
     def load_generate(self, **untracked_payload_params: Any) -> None:
@@ -183,6 +204,20 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
             self.load_generate()
         return cast(PayloadTypeT, self._payload)
 
+    def validate_payload(self, payload: Optional[PayloadTypeT] = None) -> None:
+        """
+        Validate the payload.
+
+        This method requires an implementation of _validate_payload(), which should error if the payload is invalid.
+        """
+        if payload is None:
+            if self._payload is None:
+                raise NoPayloadError("Method validate_payload() was called, but there is no payload to validate.")
+            payload = self._payload
+        self.logger.info("Validating payload...")
+        self._validate_payload(payload)
+        self.logger.info("Payload validated!")
+
     def reset_payload(self) -> None:
         self._payload = None
 
@@ -202,6 +237,21 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
 
         raise ExplainedNotImplementedError(method_name=self._generate_payload.__name__)
 
+    def _validate_payload(self, payload: PayloadTypeT) -> None:
+        """
+        Define here checks to ensure the payload is valid.
+
+        For example, a check could be that the payload is a table with five columns.
+
+        Parameters
+        ----------
+        payload    : PayloadTypeT
+            Payload to validate
+
+        Returns
+        -------
+        """
+
     def _post_load(self) -> None:
         """
         Define here extra algorithmic steps to run after loading the payload.
@@ -221,5 +271,5 @@ class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
 
     @property
     def persist_filepath(self) -> Path:
-        filename = md5(str(self.params_tree).encode()).hexdigest()
-        return self.data_dir / filename
+        params_tree_hex = md5(str({self.payload_name: self.params_tree}).encode()).hexdigest()
+        return self.data_dir / f"{self.payload_name}({params_tree_hex})"
